package ninja.thefirearchmage.games.fourtykcalculator;

import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple2;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

public class AttackScenario {
    private static final int PROBABILITY_DENOMINATOR_D6 = 6;

    // Attack attributes
    private List<Tuple2<Weapon, Integer>> weapons;
    private boolean unmodifiableHit;
    private boolean attackerWasStationary;
    private boolean attackerCharged;
    private boolean attackerInRapidFireRange;
    private boolean hasHitRollBonus;
    private boolean hasWoundRollBonus;
    private boolean hasLineOfSight;
    private boolean hasRerollHitsOnes;
    private boolean hasRerollHitsFailures;
    private boolean hasRerollHitsAny;
    private boolean hasRerollWoundsOnes;
    private boolean hasRerollWoundsFailures;
    private boolean hasRerollWoundsAny;
    private boolean fishCriticalHits;
    private boolean fishCriticalWounds;
    private int criticalHitValue;
    private int criticalWoundValue;

    // Defense attributes
    private DefensiveProfile defensiveProfile;
    private Optional<Reroll> normalSaveReroll;
    private boolean hasHitPenalty;
    private boolean hasWoundRollPenalty;
    private boolean isInCover;
    private boolean woundRollPenaltyIfStrengthIsHigher;
    private int damageReduction;

    // Result
    private AttackScenarioResolution results;

    public AttackScenario() {
        weapons = new ArrayList<>();
        unmodifiableHit = false;
        attackerWasStationary = false;
        attackerCharged = true;
        attackerInRapidFireRange = true;
        hasHitRollBonus = false;
        hasWoundRollBonus = false;
        hasLineOfSight = true;
        hasRerollHitsOnes = false;
        hasRerollHitsFailures = false;
        hasRerollHitsAny = false;
        hasRerollWoundsOnes = false;
        hasRerollWoundsFailures = false;
        hasRerollWoundsAny = false;
        fishCriticalHits = false;
        fishCriticalWounds = false;
        criticalHitValue = 6;
        criticalWoundValue = 6;

        defensiveProfile = null;
        normalSaveReroll = Optional.empty();
        hasHitPenalty = false;
        hasWoundRollPenalty = false;
        isInCover = false;
        woundRollPenaltyIfStrengthIsHigher = false;
        damageReduction = 0;

        results = new AttackScenarioResolution();
    }

    public AttackScenario addWeapon(Weapon weaponType, int numberOfWeaponsInUnit) {
        weapons.add(Tuple.of(weaponType, numberOfWeaponsInUnit));
        return this;
    }

    public AttackScenario withDefensiveProfile(DefensiveProfile defensiveProfile) {
        this.defensiveProfile = defensiveProfile;
        return this;
    }

    public void resolveAttack() {
        var resolution = new AttackScenarioResolution();
        // Filter out weapons that cannot participate in this attack
        var validWeapons = weapons.stream().filter( w -> hasLineOfSight || w._1.hasIndirectFire())
                .toList();

        // Attack step
        for(var weaponData : validWeapons) {
            Weapon weapon = weaponData._1;
            int amountOfWeapons = weaponData._2;

            // Attack probabilities
            int finalHitRollGoal = calculateHitRollGoal(weapon);
            int finalWoundRollGoal = calculateWoundRollGoal(weapon.getStrength(), defensiveProfile.getToughness(), weapon);
            var passingHits = calculatePassingHits(weapon, amountOfWeapons, finalHitRollGoal);


            int woundSuccessProbabilityNumerator = calculateProbabilityNumeratorFrom(finalWoundRollGoal);
        }
    }

    private double calculatePassingHits(Weapon weapon, int amountOfWeapons, int hitRollGoal) {
        int hitSuccessProbabilityNumerator = calculateProbabilityNumeratorFrom(hitRollGoal);
        int hitFailureProbabilityNumerator = 6 - hitSuccessProbabilityNumerator;
        double weaponAttacks = calculateWeaponAttacks(weapon);

        // Base passing hits without rerolls
        double passingHits = (weaponAttacks*amountOfWeapons) * ((double) hitSuccessProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);

        // Add hits from rerolls
        var hitRerolls = prepareHitRerolls(hitRollGoal);
        if(hitRerolls.isPresent()) {
            var rerolls = hitRerolls.get();
            passingHits += rerolls.getProbabilityNumerator() * ((double) hitSuccessProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
        }

        return passingHits;
    }

    private Optional<Reroll> prepareHitRerolls(int hitRollGoal) {
        if(hasRerollHitsOnes) {
            return Optional.of(Reroll.rerollOnOnes());
        }
        if(hasRerollHitsFailures || hasRerollHitsAny) {
            return Optional.of(Reroll.rerollFailures(hitRollGoal));
        }

        return Optional.empty();
    }

    private double calculateWeaponAttacks(Weapon weapon) {
        double weaponAttacks = weapon.getAttacks();
        if(weapon.getRapidFire().isPresent() && attackerInRapidFireRange) {
            weaponAttacks += weapon.getRapidFire().get();
        }
        results.setWeaponStat(weapon, WeaponStat.POTENTIAL_HITS_COUNT, weaponAttacks);

        // Add sustained hits
        if(weapon.getSustainedHits().isPresent()) {
            var sustainedHitsValue = weapon.getSustainedHits().get();

            double nonSustainedProbability = (double) (6 - (7 - criticalHitValue)) /PROBABILITY_DENOMINATOR_D6;
            double sustainedProbability = (double) (7 - criticalHitValue) /PROBABILITY_DENOMINATOR_D6;
            if(fishCriticalHits && hasRerollHitsAny) {
                sustainedProbability += nonSustainedProbability*sustainedProbability;
            }
            var sustainedHitsAmount = weaponAttacks * sustainedProbability * sustainedHitsValue;
            results.setWeaponStat(weapon, WeaponStat.SUSTAINED_HITS_COUNT, sustainedHitsAmount);
            weaponAttacks += sustainedHitsAmount;
        }

        return weaponAttacks;
    }

    private int calculateHitRollGoal(Weapon weapon) {
        var finalHitRollGoal = weapon.getHitValue();
        if(hasHitRollBonus) {
            finalHitRollGoal--;
        }
        if(hasHitPenalty && !unmodifiableHit) {
            finalHitRollGoal++;
        }
        if(!hasLineOfSight && weapon.hasIndirectFire()) {
            finalHitRollGoal++;
            finalHitRollGoal = Math.max(finalHitRollGoal, 4);
        }

        return finalHitRollGoal;
    }

    /**
     * Converts from a roll goal to a probability denominator of a fraction P/6. For hit and wound rolls, the die used
     * is D6.
     */
    private static int calculateProbabilityNumeratorFrom(int rollGoal) {
        // Because a roll of 1 always fails and a roll of 6 always succeeds, we modify the roll goal to be between
        // values 2-5
        var modifiedRollGoal = Math.max(rollGoal, 2);

        return 7 - modifiedRollGoal;

    }

    private int calculateWoundRollGoal(int strength, int toughness, Weapon weapon) {
        var woundRollGoal = 4;
        var finalWoundPenalty = hasWoundRollPenalty ? 1 : 0;

        if(strength > toughness) {
            finalWoundPenalty += woundRollPenaltyIfStrengthIsHigher ? 1 : 0;
            woundRollGoal = strength >= toughness*2 ? 2 : 3;
        } else if(strength < toughness) {
            woundRollGoal = toughness*2 >= strength ? 6 : 5;
        }

        var finalWoundBonus = hasWoundRollBonus ? 1 : 0;
        if(attackerCharged && weapon.hasLance() && weapon.getRangeType() == WeaponRangeType.MELEE) {
            finalWoundBonus++;
        }

        // Clever way to set either -1, 0 or 1, as the wound roll can only be modified, in the end by 1
        var finalWoundRollModifier = Integer.compare(finalWoundBonus - finalWoundPenalty, 0);

        // Having a -1 to wound rolls implies that the roll goal moves by 1
        woundRollGoal -= finalWoundRollModifier;

        return woundRollGoal;
    }
}
