package ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures;

import ninja.thefirearchmage.games.fourtykcalculator.utils.functions.CheckedCallable;
import ninja.thefirearchmage.games.fourtykcalculator.utils.functions.CheckedRunnable;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.function.*;
import java.util.stream.Collector;

/**
 * Very simple implementation of the Try object for the project.
 * Has most of the methods I finally end up using from
 * <a href="https://github.com/vavr-io/vavr/blob/master/src/main/java/io/vavr/control/Try.java">vavr Try class</a>
 */
public class Try<T> {

    private final Optional<T> result;
    private final Optional<? extends Throwable> error;

    private Try(Optional<T> result, Optional<? extends Throwable> error) {
        this.result = result;
        this.error = error;
    }
    public static Try<Void> success() {
        return new Try<>(Optional.empty(), Optional.empty());
    }
    public static <T> Try<T> success(T value) {
        return new Try<>(Optional.of(value), Optional.empty());
    }
    public static <T> Try<T> failure(Throwable error) {
        return new Try<>(Optional.empty(), Optional.of(error));
    }

    public static <T,X extends Throwable> Try<T> of(CheckedCallable<T,X> function) {
        try {
            return success(function.call());
        } catch(Throwable e) {
            return failure(e);
        }
    }
    public static <X extends Throwable> Try<Void> of(CheckedRunnable<X> function) {
        try {
            function.run();
            return success();
        } catch (Throwable e) {
            return failure(e);
        }
    }

    public T get() {
        return this.result.get();
    }
    public T orElseThrow(Function<Throwable,Throwable> errorMappingFunction) throws Throwable {
        if(isFailure()) {
            throw errorMappingFunction.apply(this.error.get());
        }

        return isVoidType() ? null : this.result.get();
    }

    /**
     * Lambdas don't play nice with checked exceptions, this allows to map the failure to a runtime exception
     * as supplied by the caller.
     * @param errorMappingFunction the checked to runtime exception mapper
     * @return the value if the try is not failing
     * @throws RuntimeException as provided by the user if the Try is a failure
     */
    public T orElseThrowRuntime(Function<Throwable,? extends RuntimeException> errorMappingFunction) throws RuntimeException {
        if(isFailure()) {
            throw errorMappingFunction.apply(this.error.get());
        }

        return isVoidType() ? null : this.result.get();
    }

    /**
     * Automatically wraps the failure into a runtime exception. Use this only if you're ok with transforming a potentially
     * checked exception into a runtime one.
     * @return the value of the Try if it is successful
     * @throws RuntimeException that wraps the original exception if the Try is a failure
     */
    public T orElseThrowRuntime() throws RuntimeException {
        if(isFailure()) {
            throw new RuntimeException(this.error.get());
        }

        return isVoidType() ? null : this.result.get();
    }

    public T orElseThrow() throws Throwable {
        if(isFailure()) {
            throw this.error.get();
        }

        return isVoidType() ? null : this.result.get();
    }

    public <R> Try<R> map(Function<T,R> mappingFunction) {
        return new Try<>(this.result.map(mappingFunction), this.error);
    }

    /**
     * If this try has failed, maps the desired type exception to the one provided in the error mapper function.
     * @param failureType the type of exception we want to map. To specify other types, call the method again
     * @param errorMapper the mapping function
     * @return the try object to keep chaining
     */
    public Try<T> mapFailure(Class<? extends Throwable> failureType, Function<Throwable,? extends Throwable> errorMapper) {
       this.error.map(e -> {
           if( failureType.isAssignableFrom(e.getClass()) ){
               return errorMapper.apply(e);
           } else {
               return e;
           }
       });

       return this;
    }

    /**
     * Executes code with the result of the current try only if this try is a success.
     * Returns a new Try with the result of the called function.
     * @param function the function to call will receive the result of the try
     * @return a new try with the result of the function, else this try.
     */
    public <R> Try<R> andThen(Function<T,R> function) {
        if(this.error.isEmpty()) {
            return Try.of(()-> function.apply(this.result.get()));
        }

        return failure(this.error.get());
    }

    /**
     * Executes code that doesn't need the result from the current try if the current try is successful, then returns a
     * new try with the result of the function.<br />
     * The typical usage will be consuming a Try&lt;Void&gt; and returning something else
     * @param function the function to call
     * @return a new Try with the result of the function, or a new try with the current failure
     * @param <R> the result type of the function
     */
    public <R> Try<R> andThen(Callable<R> function) {
        if(this.error.isEmpty()) {
            return Try.of(function::call);
        }

        return failure(this.error.get());
    }

    /**
     * Executes code only if the current try is successful. Doesn't use the try result, nor returns any value.
     * @param function the function to call
     * @return a new try from the called function or a new try with the current one's failure
     */
    public Try<Void> andThen(Runnable function) {
        if(this.error.isEmpty()) {
            return Try.of(function::run);
        }

        return failure(this.error.get());
    }

    public void ifFailure(Consumer<Throwable> errorConsumer) {
        this.error.ifPresent(errorConsumer);
    }

    public Throwable getCause() {
        return this.error.get();
    }
    public boolean isFailure() {
        return error.isPresent();
    }

    public boolean isSuccess() {
        return error.isEmpty();
    }

    /**
     * @return whether the type of result of this Try is Void
     */
    private boolean isVoidType() {
        return result.isEmpty() && error.isEmpty();
    }

    public static <R> Try<List<R>> sequence(Iterable<Try<R>> tries) {
        // Guard against empty collections coming in
        if(!tries.iterator().hasNext()) {
            return success(List.of());
        }

        boolean isVoidType = tries.iterator().next().isVoidType();
        if(isVoidType) {
            for(var item : tries) {
                if(item.isFailure()) {
                    return failure(item.getCause());
                }
            }
            return success(List.of());
        } else {
            var returns = new ArrayList<R>();
            for(var item : tries) {
                if(item.isSuccess()) {
                    returns.add(item.get());
                } else {
                    return failure(item.getCause());
                }
            }
            return success(returns);
        }
    }

    public static <R> Collector<Try<R>, List<Try<R>>, Try<List<R>>> collector() {
        return new TryCollector<>();
    }

    private static class TryCollector<R> implements Collector<Try<R>, List<Try<R>>, Try<List<R>>> {
        public TryCollector() {
        }

        @Override
        public Supplier<List<Try<R>>> supplier() {
            return ArrayList::new;
        }

        @Override
        public BiConsumer<List<Try<R>>, Try<R>> accumulator() {
            return List::add;
        }

        @Override
        public BinaryOperator<List<Try<R>>> combiner() {
            return (list1, list2) -> {
                list1.addAll(list2);
                return list1;
            };
        }

        @Override
        public Function<List<Try<R>>, Try<List<R>>> finisher() {
            return Try::sequence;
        }

        @Override
        public Set<Characteristics> characteristics() {
            return Set.of();
        }
    }
}
