package ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures;

import ninja.thefirearchmage.games.fourtykcalculator.utils.functions.Function3;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.ToString;

import java.util.function.Function;

/**
 * Data structure to hold 3 generic fields.
 * @param <T> First field type
 * @param <R> Second field type
 * @param <U> Third field type
 */
@Getter
@ToString(of = {"_1", "_2", "_3"})
@EqualsAndHashCode(of = {"_1", "_2", "_3"}, callSuper = false)
public class Tuple3<T,R,U> extends Tuple {
    final public T _1;
    final public R _2;
    final public U _3;

    Tuple3(T _1, R _2, U _3) {
        this._1 = _1;
        this._2 = _2;
        this._3 = _3;
    }

    /**
     * Maps the fields of the tuple with the given functions. Does not handle mapping errors.
     * @return a mapped object from the original tuple
     */
    public <A> A map(Function3<T,R,U,A> tuppleMapper) {
        return tuppleMapper.apply(_1, _2, _3);
    }

    /**
     * Returns a new tuple with each field being the result of applying the corresponding mapper function.
     * @param f1Mapper the mapping function to apply to the first field
     * @param f2Mapper the mapping function to apply to the second field
     * @param f3Mapper the mapping function to apply to the third field
     * @return a new tuple with the mapper functions applied
     * @param <A> the new return type for the first field
     * @param <S> the new return type for the second field
     * @param <D> the new return type for the third field
     */
    public <A,S,D> Tuple3<A,S,D> map(Function<T,A> f1Mapper,
                                     Function<R,S> f2Mapper,
                                     Function<U,D> f3Mapper) {
        return Tuple.of(f1Mapper.apply(_1), f2Mapper.apply(_2), f3Mapper.apply(_3));
    }

    /**
     * Utility function to use in reduce stream functions to accumulate tuples that contain only floats<br />
     * <code>
     *     var t1 = Tuple.of(1f,2f,3f); <br />
     *     var t2 = Tuple.of(4f, 5f, 6f); <br />
     *     var t3 = Tuple3.sum(t1, t2) <br />
     *     // t3 = (5f, 7f, 9f)
     * </code>
     * @param t1 the first float tuple
     * @param t2 the second float tuple
     * @return a new tuple with the result of adding each individual field
     */
    public static Tuple3<Float, Float, Float> sum(Tuple3<Float, Float, Float> t1,
                                                  Tuple3<Float, Float, Float> t2) {
        return Tuple.of(t1._1 + t2._1, t1._2 + t2._2, t1._3 + t2._3);
    }

    public Tuple2<R, U> remove1() {
        return Tuple.of(_2, _3);
    }
    public Tuple2<T, U> remove2() {
        return Tuple.of(_1, _3);
    }
    public Tuple2<T, R> remove3() {
        return Tuple.of(_1, _2);
    }
}
